'use strict';

const feathersHooksCommon = require('feathers-hooks-common');
const copy = require('fast-copy');
const _get = require('lodash/get.js');
const _set = require('lodash/set.js');
const _isEqual = require('lodash/isEqual.js');
const feathersUtils = require('feathers-utils');
const objectReplaceMustache = require('object-replace-mustache');
const sift = require('sift');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

const copy__default = /*#__PURE__*/_interopDefaultCompat(copy);
const _get__default = /*#__PURE__*/_interopDefaultCompat(_get);
const _set__default = /*#__PURE__*/_interopDefaultCompat(_set);
const _isEqual__default = /*#__PURE__*/_interopDefaultCompat(_isEqual);
const sift__default = /*#__PURE__*/_interopDefaultCompat(sift);

const defaultOptions = {
  skipHooks: false,
  params: void 0,
  name: "changesById",
  deleteParams: [],
  fetchBefore: false
};
const changesById = (cb, _options) => {
  const options = {
    ...defaultOptions,
    ..._options
  };
  return async (context, next) => {
    if (feathersUtils.shouldSkip("checkMulti", context)) {
      return context;
    }
    const pathBefore = getPath(options.name, true);
    if (context.type === "before" || context.type === "around") {
      const changes = await changesByIdBefore(context, options);
      if (!changes) {
        return context;
      }
      _set__default(context, pathBefore, changes);
    }
    if (next) {
      await next();
    }
    if (context.type === "after" || context.type === "around") {
      const itemsBefore = _get__default(context, pathBefore);
      const changes = await changesByIdAfter(context, itemsBefore, cb, options);
      if (!changes) {
        return context;
      }
      _set__default(context, getPath(options.name, false), changes);
    }
    return context;
  };
};
const changesByIdBefore = async (context, _options) => {
  const options = {
    ...defaultOptions,
    ..._options,
    type: "before"
  };
  let byId;
  if (context.method === "create" || !options.fetchBefore) {
    byId = {};
  } else if (context.method === "update" || context.method === "patch" || context.method === "remove") {
    byId = await getOrFindById(context, options) ?? {};
  } else {
    return [];
  }
  return byId;
};
const changesByIdAfter = async (context, itemsBefore, cb, _options) => {
  if (!itemsBefore) {
    return;
  }
  const options = {
    ...defaultOptions,
    ..._options,
    type: "after"
  };
  const items = await resultById(context, options);
  if (!items) {
    return;
  }
  const itemsBeforeOrAfter = context.method === "remove" && options.fetchBefore ? itemsBefore : items;
  const changesById2 = Object.keys(itemsBeforeOrAfter).reduce(
    (result, id) => {
      if (options.fetchBefore && (context.method !== "create" && !itemsBefore[id] || context.method !== "remove" && !items[id])) {
        throw new Error("Mismatch!");
      }
      const before = itemsBefore[id];
      const item = items[id];
      result[id] = {
        before,
        item
      };
      return result;
    },
    {}
  );
  if (cb && typeof cb === "function") {
    await cb(changesById2, context);
  }
  return changesById2;
};
const getOrFindByIdParams = async (context, options) => {
  if (context.id == null) {
    if (options.type === "before") {
      let params = copy__default(context.params);
      delete params.changesById;
      if (options?.deleteParams) {
        options.deleteParams.forEach((key) => {
          delete params[key];
        });
      }
      if (params.query?.$select) {
        delete params.query.$select;
      }
      params = Object.assign({ paginate: false }, params);
      params = typeof options.params === "function" ? await options.params(params, context) : params;
      return params;
    } else if (options.type === "after") {
      if (!options.params && !context.params.query?.$select) {
        return;
      }
      const itemOrItems = feathersHooksCommon.getItems(context);
      const idField = getIdField(context);
      if (!itemOrItems) {
        return;
      }
      const fetchedItems = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
      const ids = fetchedItems.map((x) => x && x[idField]);
      let params = {
        query: {
          [idField]: { $in: ids }
        },
        paginate: false
      };
      params = options.params ? await options.params(params, context) : params;
      return params ?? {};
    }
  } else {
    if (options.type === "after" && !options.params && !context.params.query?.$select) {
      return;
    }
    const query = Object.assign({}, context.params.query);
    delete query.$select;
    let params = Object.assign({}, context.params, { query });
    delete params.changesById;
    if (options?.deleteParams) {
      options.deleteParams.forEach((key) => {
        delete params[key];
      });
    }
    params = (typeof options.params === "function" ? await options.params(params, context) : params) ?? {};
    return params;
  }
};
const getOrFindById = async (context, _options) => {
  const options = {
    byId: true,
    ..._options
  };
  let itemOrItems;
  const idField = getIdField(context);
  const params = await getOrFindByIdParams(context, options);
  if (context.id == null) {
    const method = options.skipHooks ? "_find" : "find";
    itemOrItems = await context.service[method](params);
    itemOrItems = itemOrItems && (itemOrItems.data || itemOrItems);
  } else {
    const method = options.skipHooks ? "_get" : "get";
    itemOrItems = await context.service[method](context.id, params);
  }
  const items = !itemOrItems ? [] : Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
  if (options.byId) {
    return items.reduce((byId, item) => {
      const id = item[idField];
      byId[id] = item;
      return byId;
    }, {});
  } else {
    return items;
  }
};
const resultById = async (context, options) => {
  if (!context.result) {
    return {};
  }
  let items;
  let params = await getOrFindByIdParams(
    context,
    options
  );
  if (params) {
    const contextParams = Object.assign({}, context.params);
    delete contextParams.changesById;
    if (options?.deleteParams) {
      options.deleteParams.forEach((key) => {
        delete contextParams[key];
      });
    }
    if (_isEqual__default(params, context.params)) {
      params = null;
    }
  }
  if (context.method === "remove" || !params) {
    let itemOrItems = context.result;
    itemOrItems = Array.isArray(itemOrItems.data) ? itemOrItems.data : itemOrItems;
    items = Array.isArray(itemOrItems) ? itemOrItems : [itemOrItems];
  } else {
    items = await getOrFindById(context, {
      skipHooks: options?.skipHooks ?? false,
      byId: false,
      params: () => params,
      type: options.type
    });
  }
  const idField = context.service.id;
  return items.reduce(
    (byId, item) => {
      const id = item[idField];
      byId[id] = item;
      return byId;
    },
    {}
  );
};
const getIdField = (context) => {
  return context.service.options.id;
};
const getPath = (path, isBefore) => {
  if (isBefore) {
    if (typeof path === "string") {
      return `params.${path}.itemsBefore`;
    } else {
      return ["params", ...path, "itemsBefore"];
    }
  } else {
    if (typeof path === "string") {
      return `params.${path}`;
    } else {
      return ["params", ...path];
    }
  }
};

const trigger = (options) => {
  if (!options) {
    throw new Error("You should define subscriptions");
  }
  return async (context, next) => {
    feathersHooksCommon.checkContext(
      context,
      null,
      ["create", "update", "patch", "remove"],
      "trigger"
    );
    if (context.type === "before") {
      return await triggerBefore(context, options);
    } else if (context.type === "after") {
      return await triggerAfter(context);
    } else if (context.type === "around" && next) {
      context = await triggerBefore(context, options);
      await next();
      context = await triggerAfter(context);
      return context;
    } else {
      return context;
    }
  };
};
const triggerBefore = async (context, options) => {
  let subs = await getSubscriptions(context, options);
  if (!subs?.length) {
    return context;
  }
  if (!Array.isArray(context.data)) {
    const result = [];
    await Promise.all(
      subs.map(async (sub) => {
        if (!sub.action) {
          return;
        }
        if (sub.name && context.params.skipTrigger && (context.params.skipTrigger === sub.name || Array.isArray(context.params.skipTrigger) && context.params.skipTrigger.includes(sub.name))) {
          return;
        }
        sub.dataResolved = typeof sub.conditionsData === "function" ? await sub.conditionsData(context.data, context) : testCondition(context, context.data, sub.conditionsData ?? {});
        if (sub.dataResolved === false) {
          return;
        }
        sub.conditionsParamsResolved = typeof sub.conditionsParams === "function" ? await sub.conditionsParams(context.data, context) : testCondition(
          context,
          context.params,
          sub.conditionsParams ?? {}
        );
        if (sub.conditionsParamsResolved === false) {
          return;
        }
        result.push(sub);
      })
    );
    subs = result;
  }
  if (!subs?.length) {
    return context;
  }
  for (const sub of subs) {
    if (checkConditions(sub)) {
      continue;
    }
    sub.paramsResolved = await getOrFindByIdParams(context, {
      params: sub.params,
      deleteParams: ["trigger"],
      type: "before",
      skipHooks: false
    }) ?? {};
    sub.identifier = JSON.stringify(sub.paramsResolved);
    if (context.params.changesById?.[sub.identifier]?.itemsBefore) {
      continue;
    }
    const before = await changesByIdBefore(context, {
      skipHooks: false,
      params: () => sub.paramsResolved ? sub.paramsResolved : null,
      deleteParams: ["trigger"],
      fetchBefore: sub.fetchBefore || sub.conditionsBefore !== true
    });
    _set__default(
      context,
      ["params", "changesById", sub.identifier, "itemsBefore"],
      before
    );
  }
  setConfig(context, "subscriptions", subs);
  return context;
};
const triggerAfter = async (context) => {
  const subs = getConfig(context, "subscriptions");
  if (!subs?.length) {
    return context;
  }
  const now = /* @__PURE__ */ new Date();
  const promises = [];
  for (const sub of subs) {
    if (sub.name && context.params.skipTrigger && (context.params.skipTrigger === sub.name || Array.isArray(context.params.skipTrigger) && context.params.skipTrigger.includes(sub.name))) {
      return context;
    }
    if (checkConditions(sub)) {
      continue;
    }
    const itemsBefore = context.params.changesById?.[sub.identifier]?.itemsBefore;
    let changesById;
    if (itemsBefore) {
      changesById = await changesByIdAfter(context, itemsBefore, null, {
        name: ["changesById", sub.identifier],
        params: sub.params,
        skipHooks: false,
        deleteParams: ["trigger"],
        fetchBefore: sub.fetchBefore
      });
      _set__default(context, ["params", "changesById", sub.identifier], changesById);
    }
    changesById = context.params.changesById?.[sub.identifier];
    if (!changesById) {
      continue;
    }
    const changes = Object.values(changesById);
    const batchActionArguments = [];
    for (const change of changes) {
      const { before } = change;
      const { item } = change;
      const changeForSub = change;
      const { conditionsResult, conditionsBefore } = sub;
      let mustacheView = {
        item,
        before,
        data: context.data,
        id: context.id,
        method: context.method,
        now,
        params: context.params,
        path: context.path,
        service: context.service,
        type: context.type,
        user: context.params?.user
      };
      if (sub.view) {
        if (typeof sub.view === "function") {
          mustacheView = await sub.view(mustacheView, {
            item: changeForSub,
            items: changes,
            subscription: sub,
            subscriptions: subs,
            context
          });
        } else {
          mustacheView = Object.assign(mustacheView, sub.view);
        }
      }
      sub.resultResolved = typeof conditionsResult === "function" ? await conditionsResult({ item, before }, context) : testCondition(mustacheView, item, conditionsResult ?? {});
      if (!sub.resultResolved) {
        continue;
      }
      sub.beforeResolved = typeof conditionsBefore === "function" ? await conditionsBefore({ item, before }, context) : testCondition(mustacheView, before, conditionsBefore ?? {});
      if (!sub.beforeResolved) {
        continue;
      }
      if (isSubscriptionInBatchMode(sub)) {
        batchActionArguments.push([
          changeForSub,
          {
            subscription: sub,
            items: changes,
            context,
            view: mustacheView
          }
        ]);
      } else if (isSubscriptionNormalMode(sub)) {
        const _action = sub.action;
        const promise = _action(changeForSub, {
          subscription: sub,
          items: changes,
          context,
          view: mustacheView
        });
        if (sub.isBlocking) {
          promises.push(promise);
        }
      }
    }
    if (isSubscriptionInBatchMode(sub) && batchActionArguments.length) {
      const promise = sub.action(batchActionArguments, context);
      if (sub.isBlocking) {
        promises.push(promise);
      }
    }
  }
  await Promise.all(promises);
  return context;
};
function setConfig(context, key, val) {
  context.params.trigger = context.params.trigger || {};
  context.params.trigger[key] = val;
}
function getConfig(context, key) {
  return context.params.trigger?.[key];
}
function checkConditions(sub) {
  return !sub.conditionsBefore && !sub.conditionsData && !sub.conditionsResult;
}
const defaultSubscription = {
  name: void 0,
  action: void 0,
  conditionsBefore: true,
  conditionsData: true,
  conditionsParams: true,
  conditionsResult: true,
  dataResolved: void 0,
  beforeResolved: void 0,
  resultResolved: void 0,
  isBlocking: true,
  batchMode: false,
  method: void 0,
  params: void 0,
  service: void 0,
  view: void 0,
  paramsResolved: void 0,
  identifier: "default",
  fetchBefore: false
};
const getSubscriptions = async (context, options) => {
  const _subscriptions = typeof options === "function" ? await options(context) : options;
  if (!_subscriptions) {
    return;
  }
  const subscriptions = Array.isArray(_subscriptions) ? _subscriptions : [_subscriptions];
  const subscriptionsResolved = subscriptions.map(
    (x) => Object.assign({}, defaultSubscription, x)
  );
  const { path, method } = context;
  return subscriptionsResolved.filter((sub) => {
    if (typeof sub.service === "string" && sub.service !== path || Array.isArray(sub.service) && !sub.service.includes(path)) {
      return false;
    }
    if (typeof sub.method === "string" && sub.method !== method || Array.isArray(sub.method) && !sub.method.includes(method)) {
      return false;
    }
    return true;
  });
};
const isSubscriptionInBatchMode = (sub) => !!sub.batchMode;
const isSubscriptionNormalMode = (sub) => !sub.batchMode;
const testCondition = (mustacheView, item, conditions) => {
  if (conditions === true) {
    return true;
  }
  conditions = copy__default(conditions);
  const transformedConditions = objectReplaceMustache.replace(conditions, mustacheView);
  return sift__default(transformedConditions)(item) ? transformedConditions : false;
};

exports.changesById = changesById;
exports.changesByIdAfter = changesByIdAfter;
exports.changesByIdBefore = changesByIdBefore;
exports.getOrFindByIdParams = getOrFindByIdParams;
exports.trigger = trigger;
